;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :stella)

;;;
;;; event reading and writing. this will be implemented in terms of
;;; standard lisp streams once enough lisp vendors support stream io
;;; via clos instances and generic functions.
;;;

;(defgeneric insure-event-stream (class &optional name &rest args))
(defgeneric single-stream? (object))
(defgeneric open-event-stream (stream &optional mode &rest args))
(defgeneric close-event-stream (stream &optional mode))
(defgeneric read-event (stream))
(defgeneric write-event (event stream))
(defgeneric write-event? (event stream))
(defgeneric scale-time (event stream))
(defgeneric peek-event (stream))
(defgeneric poke-event (stream))
(defgeneric event-file-position (stream))
(defgeneric event-file-length (stream))
(defgeneric initialize-stream-for-processing (stream))
(defgeneric deinitialize-stream-for-processing (stream))
(defgeneric real-time-loop (object stream))

(defmacro openp (stream) `(slot-boundp ,stream 'stream))

(defparameter *output-stream* nil)

(defun set-output-stream (stream &optional verbose)
  (when verbose
    (tell-user (if (stringp verbose) verbose
                   "Output stream: ~S.")
               stream))
  (setf *output-stream* stream))

;;;
;;; root classes for event io.
;;;

(defclass event-stream (flags-mixin container-mixin)
  (stream
   (name :initarg name :initarg :name :accessor object-name)
   (start :initarg start :initarg :start)
   (end :initarg end :initarg :end)
   (timescale :initarg timescale :initarg :timescale)
   (direction :initarg :direction :initarg direction :initform ':output)
   (syntax :allocation :class)
   (scheduling :initarg :scheduling :initarg scheduling
               :initform ':runtime)
   (clock :initarg :clock :initarg clock)
   (initializer :initarg initializer :initarg :initializer)
   (finalizer :initarg finalizer :initarg :finalizer)
   ))

(defmethod initialize-instance :after ((stream event-stream)
                                       &rest args)
  (declare (ignore args))
  (unless (object-name stream)
    (setf (object-name stream)
          (default-stream-name stream)))
  (add-object stream .io-streams.))

(defmethod default-stream-name ((object event-stream))
  (let ((count (object-count .io-streams.)) )
    (format nil "Untitled~@[-~D~]" (and (> count 0) (1+ count)))))

(defmethod count-streams ((object string))
  (count object (container-objects .io-streams.)
         :test #'(lambda (x y) (same-name? y x))))

(defmethod count-streams ((object standard-class))
  #+pcl (setf object (class-name object))
  (count object (container-objects .io-streams.)
         :test #'(lambda (x y) (typep y x))))

(defmethod rename-ok? ((object event-stream) &optional newname)
  (declare (ignore newname))
  t)

(defmethod same-name? ((object event-stream) newname)
  (let ((name (object-name object)))
    (and (string-equal name (string newname)) name)))

(defmethod rename-object ((object event-stream) newname 
                          &optional oldname)
  (declare (ignore oldname))
  (setf (slot-value object 'name) (string newname)))

(defmethod single-stream? ((object event-stream)) nil)

(defun scheduling-mode? (stream mode)
  (eq (slot-value stream 'scheduling) mode))

;;;
;;; File streams.
;;;

(defclass event-file (event-stream)
  ((name :initarg pathname :initarg :pathname)
   (play :initarg play :initarg :play :initform ':ask)
   (element-type :initarg element-type :initarg :element-type 
                 :initform 
                 #+mcl 'base-character
                 #+(or clisp (and excl (not cltl2))) 'string-char
                 #-(or mcl clisp (and excl (not cltl2))) 'character)))

(defmethod default-stream-name ((object event-file))
  (let* ((class (class-of object))
         (count (count-streams (class-of object)))
         (syntax (slot-value object 'syntax))
         (streams (slot-value syntax 'stream-types))
         (type (second (assoc (class-name class) streams)))
         name)
    (unless type
      (error "Can't determine file type of: ~S." class))
    (setf name 
          (format nil "test~@[-~D~].~A"
                  (and (> count 0) (1+ count)) type)) 
   #+(or dos os/2 win32) (string-upcase name)
   #-(or dos os/2 win32) name))

(defmethod same-name? ((object event-file) name)
  (let* ((p1 (pathname name))
         (p2 (pathname (slot-value object 'name)))
         (d1 (pathname-directory p1))
         (d2 (pathname-directory p2))
         (n1 (pathname-name p1))
         (n2 (pathname-name p2))
         (t1 (pathname-type p1))
         (t2 (pathname-type p2)))
    (if (and d1 d2)
      (if (equal d1 d2)
        (if (and (equal n1 n2) (equal t1 t2))
          t
          nil)
        nil)
      (if (and (equal n1 n2) (equal t1 t2))
        t 
        nil))))

(defmethod print-object ((object event-file) stream)
  (format stream "#<File: ~S>" (object-name object)))

(defclass header-mixin ()
  ((header :initarg :header :initarg header)))

;;;
;;; find-stream returns a stream given a name etc.
;;;

(defmethod find-stream ((object string) &optional arg)
  (declare (ignore arg))
  ;; if the string names a syntax, find its default stream,
  ;;; otherwise find a stream with that name.
  (let ((s (find-syntax object nil)))
    (if s (find-stream s)
      (find object (container-objects .io-streams.)
            :test #'(lambda (x y) (same-name? y x))))))

;;;
;;; syntax overrides this method if it wants default streams.
;;;

(defmethod find-stream ((object syntax) &optional arg)
  (declare (ignore arg))
  nil)

(defmethod find-stream ((object standard-class) &optional arg)
  ;; find an instance of class object
  (declare (ignore arg))
  #+pcl (setf object (class-name object))
  (find-if #'(lambda (x) (typep x object))
           (container-objects .io-streams.)))

;;;
;;; main method for stream classes either finds a stream of that
;;; type and name, otherwise creates it. some streams only support
;;; single instances (midi ports) so this stream is returned
;;; regardless of its name.
;;;

(defmethod open-event-stream ((object standard-class)
                              &optional (mode ':default) &rest args)
  (let* ((name (or (pair-value 'name args) 
                   (pair-value ':name args)))
         (stream
          (if name (find-stream name)
              (let ((s (find-stream object)))
                (and s (single-stream? s) s)))))
    (unless stream
      (setf stream (apply #'make-instance object ':name name 
                           :allow-other-keys t args))
      (setf args nil)) ; null to avoid double initialization
    (apply #'open-event-stream stream mode args)))

(defmethod open-event-stream ((stream symbol) 
                              &optional (mode ':default) &rest args)
  (apply #'open-event-stream (find-class stream) mode args))

;;;
;;; main method for files opens the file, if appropriate to mode.
;;;

(defmethod open-event-stream ((stream event-file)
                              &optional (mode ':default) &rest args)
  (declare (ignore args))
  (case mode 
    ((:initialize :default) nil)
    (t
     (setf (slot-value stream 'stream)
           (open (object-name stream)
                 :direction (slot-value stream 'direction)
                 :element-type (slot-value stream 'element-type)
                 :if-exists ':supersede
                 :if-does-not-exist ':create))))
  stream)

;;;
;;; before method for all streams performs user initializations
;;; prior to opening the stream.  changes are specified using
;;; initargs, not slots, hence the MOP hackery to find the slots.
;;;

(defmethod open-event-stream :before ((object event-stream) 
                                      &optional (mode ':default)
                                      &rest args)
  (when args 
    (let ((slots (class-slots (class-of object))))
      (loop for (i v) on args by #'cddr
            for s = (find i slots :test #'member 
                          :key #'slot-definition-initargs)
            when s 
            do 
            (if (member v '(unset :unset))
              (slot-makunbound object (slot-definition-name s))
              (setf (slot-value object (slot-definition-name s)) v)))))
  (when (slot-boundp object 'clock)
    (setf (slot-value object 'clock) 
          (let ((old (clock-mode))
                (new (clock-mode (slot-value object 'clock))))
            (clock-mode old)
            new)))
  (let ((mode (slot-value object 'scheduling)))
    (unless (or (eq mode ':realtime) (eq mode ':runtime))
      (ecase mode
        ((:realtime realtime real-time :real-time)
         (unless (eq (slot-value-or-default object 'clock (clock-mode))
                     ':milliseconds)
           (warn-user
            "Warning: The current clock mode :SECONDS will result in floating ~
             point calculations that may cause GC ~
             to affect real time processing.~&"))
         (setf (slot-value object 'scheduling) ':realtime))
        ((:runtime runtime run-time :run-time)
         (setf (slot-value object 'scheduling) ':runtime)))))
  ;; really opening...
  (unless (member mode '(:initialize :default))
    ;; if system/stream clock modes dont agree, set mode and
    ;; a bit to reset it when done.
    
    (when (slot-boundp object 'initializer)
      (funcall (slot-value object 'initializer))))
  object)


;;;
;;; main method for strings finds the stream if there is one
;;; else it creates a new one if it can determine the class
;;; otherwise signals an error.
;;;

(defmethod open-event-stream ((object string)
                              &optional (mode ':default) &rest args)
  (let ((stream (find-stream object)))
    (if stream
      (apply #'open-event-stream stream mode args)
      (multiple-value-bind (syntax class)
                           (pathname-syntax object)
        (declare (ignore syntax))
        (unless class
          (error "Can't determine stream class for ~S." object))
        (apply #'open-event-stream class mode 'name object args)))))

(defmethod open-event-stream ((object syntax)
                              &optional (mode ':default) &rest args)
  (declare (ignore mode args))
  (error "Syntax ~S does not support default streams." object))

;;;
;;; close event stream closes the physical stream, if any.
;;;

(defmethod close-event-stream ((object event-stream) &optional mode)
  (declare (ignore mode))
  (when (slot-boundp object 'stream)
    (when (streamp (slot-value-or-default object 'stream nil))
      (close (slot-value object 'stream)))
    (slot-makunbound object 'stream))
  (when (slot-boundp object 'finalizer)
    (funcall (slot-value object 'finalizer))))

(defmethod close-event-stream ((object string) &optional mode)
  (when (setf object (find-stream object))
    (close-event-stream object mode)))

;;;
;;; default methods for initialization and writing do nothing.
;;;

(defmethod initialize-stream-for-processing ((stream event-stream))
  (values))

(defmethod initialize-stream-for-processing :before ((object event-stream))
  (let ((local (slot-value-or-default object 'clock (clock-mode)))) 
    (setf (slot-value object 'flags)
          (if (eql (clock-mode) local)
            (logandc2 (slot-value object 'flags) +revert-clock+)
            (progn
              (format t "must reset clock mode")
              (clock-mode local)
              (logior (slot-value object 'flags) +revert-clock+))))))

(defmethod deinitialize-stream-for-processing ((stream event-stream))
  (values))

(defmethod deinitialize-stream-for-processing :after 
           ((object event-stream))
  (when (logtest (slot-value object 'flags) +revert-clock+)
    (format t "~%resetting clock mode")
    (setf (slot-value object 'flags)
          (logandc2 (slot-value object 'flags) +revert-clock+))
    (if (eq (clock-mode) ':seconds)
      (clock-mode ':milliseconds)
      (clock-mode ':seconds))))

(defmethod write-event ((element element) (stream event-stream))
  (values))

(defmethod post-process-stream ((stream event-stream) inits)
  (declare (ignore inits))
  nil)

;;;
;;; default postprocessing asks if the file should be played
;;;

(defparameter *post-processing* ':ask)
(defparameter *post-processing-mode* ':background)

(defmethod post-process-stream ((stream event-file) inits)
  (let ((file (object-name stream))
        (play (pair-value 'play inits ':unspecified)))
    (if (eq play ':unspecified)
      (setf play (slot-value stream 'play)))
    (when (eq play ':ask)
      (setf play
            (if *command-prompting*
              (ask-y-or-n :prompt (format nil "Play file ~A? " file)
                          :default 'yes)
              nil)))
    (when play
      (apply #'play-using-syntax 
             (slot-value stream 'syntax) file inits))))

;;;
;;; write-event is the main function for output processing.  this method on
;;; containers is our main processing loop.  it calls select-element to get
;;; the next element and then write-event to output the object. each type of 
;;; object must implement a method for the output stream(s) it wants to
;;; support. write-event performs two possible main loops, based on whether
;;; the stream has a start or end time associated with it. if it does,
;;; each selected object is checked to see if it should be output, and 
;;; then is "left timeshifted" by the start time amount. if no start or end
;;; is specified then we optimize out the additional checks and processing.
;;;

(defmethod write-event ((object container) (stream event-stream))
  (score-reset object)
  (if (scheduling-mode? stream ':realtime)
    (process-run-function "Event Scheduler"
                          #'real-time-loop object stream)
    (let ((ok? nil)
          (time (slot-value object 'time))
          event)
      (unwind-protect 
        (if (or (slot-boundp stream 'start) 
                (slot-boundp stream 'end))
          (loop while (setf event (score-select object time))
                when (write-event? event stream)
                do (shift-time event stream)
                (write-event event stream)
                do (score-unset event)
                finally (setf ok? t))
          (loop while (setf event (score-select object time))
                do (write-event event stream)
                (score-unset event)
                finally (setf ok? t)))
        (unless ok?
          (warn-user "Processing of ~S unexpectedly terminated.~&~
                      Performing recursive unset..." object)
          (deallocate-all-objects *scheduler-resource*))
        (score-unset object (not ok?)))
      object)))

;;; this is overridden by MIDI ports to use MIDI time services.

(defmethod real-time-loop ((object container) (stream event-stream))
  (macrolet ((clock-time ()
               `(if (eq *clock-mode* :milliseconds)
                  (get-internal-real-time)
                  (/ (get-internal-real-time) 
                     ,(coerce internal-time-units-per-second
                              'float)))))
    (let* ((*merge* object)
           (okay nil)
           (time (slot-value object 'time))
           (base (clock-time))
           (this (+ base (slot-value (peek-object object) 'time)))
           event)
      (declare (dynamic-extent *merge*))
      (unwind-protect 
        (loop do
              (process-wait "Waiting"
                            #'(lambda (m) (> (clock-time) m))
                            this)
              while (setf event (score-select object time))
              do
              (write-event event stream)
              (setf this (slot-value event 'time))
              (score-unset event)
              finally (setf okay t))
        (if okay
          (progn
            (score-unset object nil)
            (close-event-stream stream nil)
            (post-process-stream stream nil))
          (progn
            (warn-user "Processing of ~S unexpectedly terminated.~&~
                        Performing recursive unset..." object)
            (deallocate-all-objects *scheduler-resource*)
            (score-unset object t)
            (close-event-stream stream ':error)))))))

;;;
;;; :around method for merges supports dynamic queuing during element
;;; writing. this is used by midi-note elements to schedule a future
;;; noteOff each time a note is written to the stream.
;;;

(defmethod write-event :around ((object merge) (stream event-stream))
  (let ((*merge* object)) 
    (call-next-method)))



;;;
;;; write-event? will be called on an object just before it is to be output
;;; if the stream has a start or end time associated with it. 
;;; it should return t if the event should be written, otherwise nil.
;;;

(defmethod write-event? ((object timed-object) (stream event-stream))
  (if (slot-boundp stream 'start)
      (if (< (slot-value object 'time) ;0
             (slot-value stream 'start))  
         nil
        (if (slot-boundp stream 'end)
            (if (> (slot-value object 'time) (slot-value stream 'end))
                nil
              t)
          t))
    (if (slot-boundp stream 'end)
        (if (> (slot-value object 'time) (slot-value stream 'end))
            nil
          t)
      t)))

;;;
;;; shift object's time left just before output. some objects, like
;;; midi-releases, shouldn't be time shifted, hence the generic function.
;;;

(defmethod shift-time ((object timed-object) (stream event-stream))
  (when (slot-boundp stream 'start)
    (setf (slot-value object 'time)
      (- (slot-value object 'time) (slot-value stream 'start)))))

;;;
;;; streams and syntax
;;;




(defgeneric open-listener (syntax &rest args))
(defgeneric close-listener (syntax &optional mode))
(defgeneric syntax-default-pathname (syntax stream))
(defgeneric play-using-syntax (syntax file &rest args))
(defgeneric import-using-syntax (syntax file &rest args))
(defgeneric load-using-syntax (syntax file &rest args))

(defmethod syntax-default-pathname ((syntax syntax) (stream event-file))
  (slot-value syntax 'pathname))

(defmethod open-listener ((syntax syntax) &rest args)
  (declare (ignore args))
  nil)

(defmethod close-listener ((syntax syntax) &optional  mode)
  (declare (ignore mode))
  nil)

;;;
;;;
;;;

(defun keyword-prefix (slot &optional (space t))
  (if space (format nil ":~(~A~) " slot)
    (format nil ":~(~A~) " slot)))

(defun &rest-printer (values stream)
  ;; initial argument delimiter is already printed.
  (loop while values
        do (prin1 (pop values) stream)
           (and values (write-char #\space stream))))

(defmacro formatting-slots ((object stream &rest keys) &rest slots)
  (apply #'formatting-slots-internal object stream slots keys))

(defun formatting-slots-internal (object stream specs 
                              &key (printer 'princ) (print-if t)
                                   eol prefix suffix format filter decimals
                                   (delimiter '#\space) preamble postamble
                                   constructor default)
  (let (($stream$ (gensym))
        ($object$ (gensym))
        (forms ()))
    (setf forms
      (loop with del = nil 
            for spec in specs 
            when (eq spec '&optional)
            do (setf prefix nil print-if :bound )
            else
            when (eq spec '&key)
            do (setf prefix 'keyword-prefix print-if :bound default nil)
            else
            when (eq spec '&rest)
            do (setf prefix nil printer '&rest-printer print-if :bound 
                     default nil)
            else
            collect 
            (if (consp spec)
                (apply 
                  #'(lambda (slot &key (printer printer)
                                       (print-if print-if)
                                       (eol nil)
                                       (prefix prefix)
                                       (delimiter delimiter dp)
                                       (suffix suffix)
                                       (filter filter)
                                       (decimals decimals)
                                       (format format)
                                       (constructor constructor)
                                       (default default)
                                       )
                       (slot-format-form $object$ slot $stream$
                                         print-if printer format
                                         (if dp delimiter
                                           (if del delimiter nil))                            
                                         prefix suffix eol filter
                                         constructor default decimals))
                  spec)
               (slot-format-form $object$ spec $stream$
                                  print-if printer format 
                                  (if del delimiter nil)
                                  prefix suffix nil filter
                                  constructor default decimals))
           and do (setf del t)))
    (when preamble
      (let ((printer (select-printer preamble)))
        (push `(,printer ,preamble ,$stream$) forms)))
    (when postamble
      (let ((printer (select-printer postamble)))
        (nconc forms (list `(,printer ,postamble ,$stream$)))))
    (when eol
      (nconc forms (list `(terpri ,$stream$))))
    `(let (,@ (if stream `((,$stream$ ,stream)) nil)
           ,@ (if object `((,$object$ ,object)) nil))
        ,.forms)))


(defun slot-format-form (object slot stream print-if printer 
                                format delimiter prefix suffix eol 
                                filter constructor default decimals)
  (let (value check-value check-bound forms)
    (cond (default 
           (setf value `(if (slot-boundp ,object ',slot)
                            (slot-value ,object ',slot)
                          ,default)))
          ((member print-if '(:value :valued))
           (setf value (gensym) check-value t check-bound t))
          ((eq print-if ':bound)
           (setf value `(slot-value ,object ',slot) check-bound t))
          ((member print-if '(t :always))
           (setf value `(slot-value ,object ',slot))
           (setf check-bound (eq print-if :always)))
          (t
           (error "~S is not a legal :PRINT-IF value." print-if)))
    ;; :decimal value becomes filter
    (when decimals
      (when filter (error "Exclusive keywords: decimal, filter"))
      (unless (and (integerp decimals) (>= decimals 0))
        (error "decimals: ~S not >= 0."
               decimals))
      (setf filter
        (if (= 0 decimals)
            'floor
          #+akcl
          (let ((str (format nil "~~,~df" decimals)))
            `(lambda (x) (format nil ,str x)))
          #-akcl
          `(lambda (x)
             (/ (round x ,(expt 10.0 (- decimals)))
                    ,(expt 10.0 decimals))))))
    (setf forms
      (cond (constructor 
             (when (or printer format delimiter prefix suffix eol)
               (error ":constructor excludes other keywords."))
             (list (list (funcall (coerce constructor 'function)
                                  slot) value)))
            (printer
             (list `(,printer ,(if filter (list filter value) value) 
                     ,stream)))
            ))
    (when format
      (ecase format
        (:quoted (push `(write-char #\' ,stream) forms))
        (:careful (push (let ((v (gensym)))
                          `(let ((,v (slot-value ,object ',slot)))
                             (and (or (consp ,v) 
                                      (and (symbolp ,v) (not (boundp ,v))))
                                  (write-char #\' ,stream))))
                      forms))
        (:string (push `(write-char #\" ,stream) forms)
                 (nconc forms (list `(write-char #\" ,stream))))))

    (when delimiter  
     (when (and (symbolp delimiter) (fboundp delimiter))
       (setf delimiter (funcall delimiter slot))))

    (when prefix   ; prefix is char, string or fn to call on slot
      (when (or (and (symbolp prefix) (fboundp prefix))
                (and (consp prefix) (eq (first prefix) 'lambda)
                     (setf prefix (coerce prefix 'function))))
        (setf prefix (funcall prefix slot)))
        ;; combine delimiter and prefix if possible.
        (when delimiter
          (setf prefix (concatenate 'string (string delimiter) 
                                    (string prefix)))
          (setf delimiter nil))
       (let ((printer (select-printer prefix)))
        (push `(,printer ,prefix ,stream) forms)))
    (when suffix   ; suffix is char, string or fn to call on slot
      (when (and (symbolp suffix) (fboundp suffix))
        (setf suffix (funcall suffix slot)))
      (let ((printer (select-printer suffix)))
        (nconc forms (list `(,printer ,suffix ,stream)))))
   (when delimiter  
     (let ((printer (select-printer delimiter)))
       (push `(,printer ,delimiter ,stream) forms)))
   (when eol
      (nconc forms (list `(terpri ,stream))))
    (when check-value
      (setf forms
        `((let ((,value (slot-value ,object ',slot)))
            (when ,value ,. forms)))))
    (if check-bound
        (if (eq print-if ':always) 
            `(cond ((slot-boundp ,object ',slot) ,. forms)
                   (t ; maybe print delimiter and print unset
                     ,@(if delimiter 
                           `((,(select-printer delimiter)
                              ,delimiter ,stream)))
                      (,printer ,+slot-unset+ ,stream)))
         `(when (slot-boundp ,object ',slot)
             ,. forms))
       `(progn ,.forms))))

(defun select-printer (thing)
  (cond ((stringp thing) 
         'write-string)
        ((characterp thing)
         'write-char)
        (t
         (error "~S is not a string or character." thing))))

(defmacro collecting-slots (object &body slots)
  (let ((vals (gensym))
        specs)
    (flet ((fn1 (s) s `(lambda (x) (push x ,vals)))
           (fn2 (s) 
             `(lambda (val) ; push key first because of nreverse!!!
                (push ,(intern (string s) :keyword) ,vals)
                (push val ,vals))))      
       (setf specs
            (loop with if = t and fn = #'fn1
                  for s in slots
                  when (eq s '&optional)
                  do (setf if ':bound)
                  else
                  when (eq s '&key)
                  do (setf if ':bound fn #'fn2)
                  else
                  collect (if (listp s) 
                            (append s (list :print-if if :constructor fn))
                            (list s :print-if if :constructor fn))))
      `(let ((,vals ()))
         ,(formatting-slots-internal object nil specs :printer nil 
                                     :delimiter nil)
         (nreverse ,vals)))))


#|
;;; tests
(pprint (macroexpand '(formatting-slots (obj stream :default 0 
                                         :delimiter #\space)
                        foo bar (baz :filter note))))

(pprint (macroexpand '(formatting-slots (obj stream :constructor mk-make-sender
                                         :print-if :bound :printer nil)
                        freq amp )))
(pprint (macroexpand '(collecting-slots obj
                        foo &key bar (baz :default 12))))
(pprint (macroexpand '(collecting-slots o rhythm &key duration freq amp)))
;i1 1.5 1.43 256.734 1 2 45.999 5
(pprint (macroexpand 
'(formatting-slots (object (slot-value stream 'stream) 
                        :eol t :decimals 2)
     (name :decimals nil)
     time dur freq amp ind loc dis (int :decimals 0))))
|#


